# cmultiply.pyx

# Import Cython and types from standard Python
import numpy as np
cimport numpy as np
from libc.stdio cimport sprintf
import numpoly
from typing import List


cdef void cmultiply_cdef(
        np.ndarray[np.uint32_t, ndim=2] expons1,
        np.ndarray[np.uint32_t, ndim=2] expons2,
        list coeffs1,
        list coeffs2,
        int offset,
        np.ndarray out,
):
    cdef Py_ssize_t i, j, k

    # Declare variables for string representation of polynomials
    cdef set seen = set()
    cdef char key[256]
    cdef Py_ssize_t key_len
    cdef str key_str

    for i in range(expons1.shape[0]):
        for j in range(expons2.shape[0]):
            key_len = 0
            for k in range(expons1.shape[1]):
                key_len += sprintf(key + key_len, "%c", expons1[i, k] + expons2[j, k] + offset)

            key_str = key[:key_len].decode('utf-8')
            if key_str in seen:
                numpoly.cadd_values((coeffs1[i] * coeffs2[j]).ravel(), key_str, out)
            else:
                numpoly.cset_values((coeffs1[i] * coeffs2[j]).ravel(), key_str, out)
                seen.add(key_str)


def cmultiply(
        np.ndarray[np.uint32_t, ndim=2] expons1,
        np.ndarray[np.uint32_t, ndim=2] expons2,
        list coeffs1,
        list coeffs2,
        int offset,
        np.ndarray out,
):
    cmultiply_cdef(expons1, expons2, coeffs1, coeffs2, offset, out)



"""Return the minimum of an array or minimum along an axis."""
from __future__ import annotations
from typing import Any, Optional, Sequence, Union

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements

if tuple(int(v) for v in numpy.__version__.split(".")) >= (1, 25, 0):
    impls = [numpy.amin, numpy.min, min]
else:
    impls = [numpy.min, min]


@implements(*impls)
def amin(
    a: PolyLike,
    axis: Union[None, int, Sequence[int]] = None,
    out: Optional[ndpoly] = None,
    **kwargs: Any,
) -> ndpoly:
    """
    Return the minimum of an array or minimum along an axis.

    Args:
        a:
            Input data.
        axis:
            Axis or axes along which to operate.  By default, flattened input
            is used. If this is a tuple of ints, the minimum is selected over
            multiple axes, instead of a single axis or all the axes as before.
        out:
            Alternative output array in which to place the result. Must be of
            the same shape and buffer length as the expected output.
        keepdims:
            If this is set to True, the axes which are reduced are left in the
            result as dimensions with size one. With this option, the result
            will broadcast correctly against the input array.

            If the default value is passed, then `keepdims` will not be passed
            through to the `amax` method of sub-classes of `ndarray`, however
            any non-default value will be.  If the sub-class' method does not
            implement `keepdims` any exceptions will be raised.
        initial:
            The minimum value of an output element. Must be present to allow
            computation on empty slice.
        where:
            Elements to compare for the maximum.

    Return:
        Minimum of `a`. If `axis` is None, the result is a scalar value.
        If `axis` is given, the result is an array of dimension ``a.ndim-1``.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.amin([13, 7])
        polynomial(7)
        >>> numpoly.amin([1, q0, q0**2, q1])
        polynomial(1)
        >>> numpoly.amin([q0, q1, q0**2])
        polynomial(q0)
        >>> numpoly.amin([[3*q0**2, q0**2],
        ...               [2*q0**2, 4*q0**2]], axis=1)
        polynomial([q0**2, 2*q0**2])

    """
    del out
    poly = numpoly.aspolynomial(a)
    options = numpoly.get_options()
    proxy = numpoly.sortable_proxy(
        poly, graded=options["sort_graded"], reverse=options["sort_reverse"]
    )
    indices = numpy.amin(proxy, axis=axis, **kwargs)
    out = poly[numpy.isin(proxy, indices)]
    out = out[numpy.argsort(indices.ravel())]
    return numpoly.reshape(out, indices.shape)



"""Test whether any array element along a given axis evaluates to True."""
from __future__ import annotations
from typing import Any, Optional, Sequence, Union

import numpy
import numpy.typing
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


@implements(numpy.any)
def any(
    a: PolyLike,
    axis: Union[None, int, Sequence[int]] = None,
    out: Optional[numpy.ndarray] = None,
    keepdims: bool = False,
    **kwargs: Any,
) -> Optional[numpy.ndarray]:
    """
    Test whether any array element along a given axis evaluates to True.

    Returns single boolean unless `axis` is not ``None``

    Args:
        a:
            Input array or object that can be converted to an array.
        axis:
            Axis or axes along which a logical OR reduction is performed. The
            default (`axis` = `None`) is to perform a logical OR over all the
            dimensions of the input array. `axis` may be negative, in which
            case it counts from the last to the first axis. If this is a tuple
            of ints, a reduction is performed on multiple axes, instead of
            a single axis or all the axes as before.
        out:
            Alternate output array in which to place the result.  It must have
            the same shape as the expected output and its type is preserved
            (e.g., if it is of type float, then it will remain so, returning
            1.0 for True and 0.0 for False, regardless of the type of `a`).
        keepdims:
            If this is set to True, the axes which are reduced are left in the
            result as dimensions with size one. With this option, the result
            will broadcast correctly against the input array.

    Return:
        A new boolean or `ndarray` is returned unless `out` is specified,
        in which case a reference to `out` is returned.

    Example:
        >>> q0 = numpoly.variable()
        >>> numpoly.any(q0)
        np.True_
        >>> numpoly.any(0*q0)
        np.False_
        >>> numpoly.any([1, q0, 0])
        np.True_
        >>> numpoly.any([[True*q0, False], [True, True]], axis=0)
        array([ True,  True])

    """
    a = numpoly.aspolynomial(a)
    coefficients = numpy.any(numpy.asarray(a.coefficients), axis=0)
    index = numpy.asarray(coefficients, dtype=bool)
    return numpy.any(index, axis=axis, out=out, keepdims=keepdims)



"""Testing functions used for numpy compatible."""
from __future__ import division
from pytest import raises
import numpy
from numpoly import polynomial
import numpoly

X, Y = numpoly.variable(2)


def assert_equal(results, reference, c_contiguous=None,
                 f_contiguous=None, type_=None):
    """
    Assert that a return value for a function is the same as reference.

    Checks types, values, datatype, shape, C- and F-contiguous-ness.

    Args:
        results (numpy.ndarray, numpoly.ndpoly):
            The results to check are correct.
        reference (numpy.ndarray, numpoly.ndpoly):
            The reference the results is checked against. Input will be cast to
            the same type as `results`.
        c_contiguous (Optional[bool]):
            Check if `results` has correct `C_CONTIGUOUS` flag. Checked against
            `reference` if not provided.
        f_contiguous (Optional[bool]):
            Check if `results` has correct `F_CONTIGUOUS` flag. Checked against
            `reference` if not provided.
        type_ (Optional[type]):
            Check if `results` is correct type using `isinstance`. If not
            provided, results are checked to be legal numpy or numpoly type.

    """
    if type_ is None:
        assert isinstance(
            results, (bool, numpy.bool_, numpy.number, numpy.ndarray)), (
                f"unrecognized results type: {results}")
    else:
        assert isinstance(results, type_), (
            f"invalid results type: {results} != {type_}")
    if isinstance(results, numpoly.ndpoly):
        reference = numpoly.aspolynomial(reference)
    else:
        results = numpy.asarray(results)
        reference = numpy.asarray(reference)

    assert results.shape == reference.shape, (
        f"shape mismatch: {results} != {reference}")
    assert results.dtype == reference.dtype, (
        f"dtype mismatch: {results} != {reference}")

    if not isinstance(results, numpoly.ndpoly):
        assert numpy.allclose(results, reference), (
            f"value mismatch: {results} != {reference}")
    elif results.shape:
        assert numpy.all(results == reference), (
            f"value mismatch: {results} != {reference}")
    else:
        assert results == reference, (
            f"value mismatch: {results} != {reference}")

    if c_contiguous is None:
        c_contiguous = reference.flags["C_CONTIGUOUS"]
    assert results.flags["C_CONTIGUOUS"] == c_contiguous, (
        f"c_contiguous mismatch: {results} != {reference}")
    if f_contiguous is None:
        f_contiguous = reference.flags["F_CONTIGUOUS"]
    assert results.flags["F_CONTIGUOUS"] == f_contiguous, (
        f"f_contiguous mismatch: {results} != {reference}")


def test_absolute(interface):
    """Tests for numpoly.absolute."""
    assert_equal(X, abs(-X))
    assert_equal(abs(X), abs(-X))
    assert_equal(interface.abs(polynomial([X-Y, Y-4])), [X+Y, Y+4])


def test_add(interface):
    """Tests for numpoly.add."""
    assert_equal(interface.add(X, 3), 3+X)
    assert_equal(interface.add(polynomial([1, X, Y]), 4), [5, 4+X, 4+Y])
    assert_equal(interface.add(polynomial([0, X]), polynomial([Y, 0])), [Y, X])
    assert_equal(interface.add(polynomial(
        [[1, X], [Y, X*Y]]), [2, X]), [[3, 2*X], [2+Y, X+X*Y]])


def test_any(interface):
    """Tests for numpoly.any."""
    poly = polynomial([[0, Y], [0, 0]])
    assert_equal(interface.any(poly), True)
    assert_equal(interface.any(poly, axis=0), [False, True])
    assert_equal(
        interface.any(poly, axis=-1, keepdims=True), [[True], [False]])


def test_all(interface):
    """Tests for numpoly.all."""
    poly = polynomial([[0, Y], [X, 1]])
    assert_equal(interface.all(poly), False)
    assert_equal(interface.all(poly, axis=0), [False, True])
    assert_equal(
        interface.all(poly, axis=-1, keepdims=True), [[False], [True]])


def test_allclose(func_interface):
    """Tests for numpoly.allclose."""
    poly1 = numpoly.polynomial([1e10*X, 1e-7])
    poly2 = numpoly.polynomial([1.00001e10*X, 1e-8])
    assert_equal(func_interface.allclose(poly1, poly2), False, type_=bool)
    poly1 = numpoly.polynomial([1e10*X, 1e-8])
    poly2 = numpoly.polynomial([1.00001e10*X, 1e-9])
    assert_equal(func_interface.allclose(poly1, poly2), True, type_=bool)
    poly2 = numpoly.polynomial([1e10*Y, 1e-8])
    assert_equal(func_interface.allclose(poly1, poly2), False, type_=bool)


def test_amax(func_interface):
    """Tests for numpoly.amax."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(func_interface.amax(poly), X**4)
    assert_equal(func_interface.amax(poly, axis=0), [X**4, Y**2, X**2, Y**2])
    assert_equal(func_interface.amax(poly, axis=1), [X**2, X**2, Y**2, X**4])
    assert_equal(func_interface.amax(poly.reshape(2, 2, 2, 2), axis=(0, 1)),
                 [[X**4, Y**2], [X**2, Y**2]])


def test_amin(func_interface):
    """Tests for numpoly.amin."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(func_interface.amin(poly), 1)
    assert_equal(func_interface.amin(poly, axis=0), [1, 3, 2, X])
    assert_equal(func_interface.amin(poly, axis=1), [1, 2, 3, Y])
    assert_equal(func_interface.amin(poly.reshape(2, 2, 2, 2), axis=(0, 1)),
                 [[1, 3], [2, X]])


def test_argmax(func_interface):
    """Tests for numpoly.argmax."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(func_interface.argmax(poly), 13)
    assert_equal(func_interface.argmax(poly, axis=0), [3, 3, 1, 0])
    assert_equal(func_interface.argmax(poly, axis=1), [3, 2, 1, 1])


def test_argmin(func_interface):
    """Tests for numpoly.argmin."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(func_interface.argmin(poly), 0)
    assert_equal(func_interface.argmin(poly, axis=0), [0, 0, 2, 1])
    assert_equal(func_interface.argmin(poly, axis=1), [0, 3, 2, 2])


def test_apply_along_axis(func_interface):
    """Tests for numpoly.apply_along_axis."""
    np_array = numpy.arange(9, dtype=int).reshape(3, 3)
    assert_equal(
        func_interface.apply_along_axis(numpy.sum, 0, np_array), [9, 12, 15])
    assert_equal(
        func_interface.apply_along_axis(numpy.sum, 1, np_array), [3, 12, 21])
    poly1 = numpoly.polynomial([[X, X, X], [Y, Y, Y], [1, 2, 3]])
    assert_equal(
        func_interface.apply_along_axis(numpoly.sum, 0, poly1),
        [X+Y+1, X+Y+2, X+Y+3])
    assert_equal(
        func_interface.apply_along_axis(numpoly.sum, 1, poly1), [3*X, 3*Y, 6])


def test_apply_over_axes(func_interface):
    """Tests for numpoly.apply_over_axes."""
    np_array = numpy.arange(9).reshape(3, 3)
    assert_equal(
        func_interface.apply_over_axes(numpy.sum, np_array, 0), [[9, 12, 15]])
    assert_equal(
        func_interface.apply_over_axes(numpy.sum, np_array, 1),
        [[3], [12], [21]])
    poly1 = numpoly.polynomial([[X, X, X], [Y, Y, Y], [1, 2, 3]])
    assert_equal(
        func_interface.apply_over_axes(
            numpoly.sum, poly1, 0), [[X+Y+1, X+Y+2, X+Y+3]])
    assert_equal(
        func_interface.apply_over_axes(
            numpoly.sum, poly1, 1), [[3*X], [3*Y], [6]])


def test_around(interface):
    """Tests for numpoly.around."""
    poly = 123.45*X+Y
    assert_equal(interface.round(poly), 123.*X+Y)
    assert_equal(interface.round(poly, decimals=1), 123.4*X+Y)
    assert_equal(interface.round(poly, decimals=-2), 100.*X)
    out = 5.*X+6.*Y
    interface.round(poly, decimals=1, out=out)
    assert_equal(out, 123.4*X+Y)


def test_array_repr(func_interface):
    """Tests for numpoly.array_repr."""
    assert repr(polynomial([])).startswith("polynomial([], dtype=")
    assert repr(4+6*X**2) == "polynomial(6*q0**2+4)"
    assert func_interface.array_repr(4+6*X**2) == "polynomial(6*q0**2+4)"
    assert (repr(polynomial([1., -5*X, 3-X**2])) ==
            "polynomial([1.0, -5.0*q0, -q0**2+3.0])")
    assert (func_interface.array_repr(polynomial([1., -5*X, 3-X**2])) ==
            "polynomial([1.0, -5.0*q0, -q0**2+3.0])")
    assert repr(polynomial([[[1, 2], [5, Y]]])) == """\
polynomial([[[1, 2],
             [5, q1]]])"""
    assert func_interface.array_repr(polynomial([[[1, 2], [5, Y]]])) == """\
polynomial([[[1, 2],
             [5, q1]]])"""


def test_array_split(func_interface):
    """Tests for numpoly.array_split."""
    test_split(func_interface)
    poly = numpoly.polynomial([[1, X, X**2], [X+Y, Y, Y]])
    part1, part2 = func_interface.array_split(poly, 2, axis=1)
    assert_equal(part1, [[1, X], [X+Y, Y]])
    assert_equal(part2, [[X**2], [Y]])


def test_array_str(func_interface):
    """Tests for numpoly.array_str."""
    assert str(polynomial([])).startswith("[]")
    assert str(4+6*X**2) == "6*q0**2+4"
    assert func_interface.array_str(4+6*X**2) == "6*q0**2+4"
    assert str(polynomial([1., -5*X, 3-X**2])) == "[1.0 -5.0*q0 -q0**2+3.0]"
    assert func_interface.array_str(
        polynomial([1., -5*X, 3-X**2])) == "[1.0 -5.0*q0 -q0**2+3.0]"
    assert str(polynomial([[[1, 2], [5, Y]]])) == """\
[[[1 2]
  [5 q1]]]"""
    assert func_interface.array_str(polynomial([[[1, 2], [5, Y]]])) == """\
[[[1 2]
  [5 q1]]]"""


def test_atleast_1d(func_interface):
    """Tests for numpoly.atleast_1d."""
    polys = [X, [X], [[X]], [[[X]]]]
    results = func_interface.atleast_1d(*polys)
    assert isinstance(results, list)
    assert_equal(results[0], [X])
    assert_equal(results[1], [X])
    assert_equal(results[2], [[X]])
    assert_equal(results[3], [[[X]]])


def test_atleast_2d(func_interface):
    """Tests for numpoly.atleast_2d."""
    polys = [X, [X], [[X]], [[[X]]]]
    results = func_interface.atleast_2d(*polys)
    assert isinstance(results, list)
    assert_equal(results[0], [[X]])
    assert_equal(results[1], [[X]])
    assert_equal(results[2], [[X]])
    assert_equal(results[3], [[[X]]])


def test_atleast_3d(func_interface):
    """Tests for numpoly.atleast_3d."""
    polys = [X, [X], [[X]], [[[X]]]]
    results = func_interface.atleast_3d(*polys)
    assert isinstance(results, list)
    assert_equal(results[0], [[[X]]])
    assert_equal(results[1], [[[X]]])
    assert_equal(results[2], [[[X]]])
    assert_equal(results[3], [[[X]]])


def test_broadcast_array(func_interface):
    """Tests for numpoly.broadcast_array."""
    polys = [X, [[Y, 1]], [[X], [Y]], [[X, 1], [Y, 2]]]
    results = func_interface.broadcast_arrays(*polys)
    assert isinstance(results, list)
    assert len(results) == len(polys)
    assert all(result.shape == (2, 2) for result in results)
    assert_equal(results[0], [[X, X], [X, X]])
    assert_equal(results[1], [[Y, 1], [Y, 1]])
    assert_equal(results[2], [[X, X], [Y, Y]])
    assert_equal(results[3], [[X, 1], [Y, 2]])


def test_ceil(func_interface):
    """Tests for numpoly.ceil."""
    poly = polynomial([-1.7*X, X-1.5, -0.2, 3.2+1.5*X, 1.7, 2.0])
    assert_equal(func_interface.ceil(poly),
                 [-X, -1.0+X, 0.0, 4.0+2.0*X, 2.0, 2.0])


def test_common_type(func_interface):
    """Tests for numpoly.common_type."""
    assert func_interface.common_type(
        numpy.array(2, dtype=numpy.float32)) == numpy.float32
    assert func_interface.common_type(X) == numpy.float64
    assert func_interface.common_type(
        numpy.arange(3), 1j*X, 45) == numpy.complex128


def test_concatenate(func_interface):
    """Tests for numpoly.concatenate."""
    poly1 = polynomial([[0, Y], [X, 1]])
    assert_equal(func_interface.concatenate([poly1, poly1]),
                 [[0, Y], [X, 1], [0, Y], [X, 1]])
    assert_equal(func_interface.concatenate([poly1, [[X*Y, 1]]], 0),
                 [[0, Y], [X, 1], [X*Y, 1]])
    assert_equal(func_interface.concatenate([poly1, [[X*Y], [1]]], 1),
                 [[0, Y, X*Y], [X, 1, 1]])
    assert_equal(func_interface.concatenate([poly1, poly1], 1),
                 [[0, Y, 0, Y], [X, 1, X, 1]])


def test_copyto(func_interface):
    """Tests for numpoly.copyto."""
    poly = numpoly.polynomial([1, X, Y])
    poly_ref = numpoly.polynomial([1, X, Y])
    with raises(ValueError):
        func_interface.copyto(poly.values, poly_ref, casting="safe")
    with raises(ValueError):
        numpoly.copyto(poly.values, [1, 2, 3], casting="safe")
    with raises(ValueError):
        numpoly.copyto(X, Y, casting="unsafe")
    func_interface.copyto(poly, X)
    assert_equal(poly, [X, X, X])
    func_interface.copyto(poly.values, poly_ref, casting="unsafe")
    assert_equal(poly, poly_ref)
    func_interface.copyto(poly, 4)
    assert_equal(poly, [4, 4, 4])
    func_interface.copyto(poly.values, poly_ref.values, casting="unsafe")
    assert_equal(poly, poly_ref)
    poly = numpoly.polynomial([1, 2, 3])
    func_interface.copyto(poly, [3, 2, 1], casting="unsafe")
    assert_equal(poly, [3, 2, 1])
    func_interface.copyto(
        poly.values, numpoly.polynomial([1, 2, 3]), casting="unsafe")
    assert_equal(poly, [1, 2, 3])
    out = numpy.zeros(3, dtype=float)
    numpoly.copyto(out, poly, casting="unsafe")
    assert_equal(out, [1., 2., 3.])


def test_count_nonzero(func_interface):
    """Tests for numpoly.count_nonzero."""
    poly1 = polynomial([[0, Y], [X, 1]])
    poly2 = polynomial([[0, Y, X, 0, 0], [3, 0, 0, 2, 19]])
    assert_equal(func_interface.count_nonzero(poly1), 3, type_=int)
    assert_equal(func_interface.count_nonzero(poly1, axis=0), [1, 2])
    assert_equal(func_interface.count_nonzero(poly2, axis=0), [1, 1, 1, 1, 1])
    assert_equal(func_interface.count_nonzero(X), 1, type_=int)


def test_cumsum(interface):
    """Tests for numpoly.cumsum."""
    poly1 = polynomial([[0, Y], [X, 1]])
    assert_equal(interface.cumsum(poly1), [0, Y, X+Y, 1+X+Y])
    assert_equal(interface.cumsum(poly1, axis=0), [[0, Y], [X, Y+1]])
    assert_equal(interface.cumsum(poly1, axis=1), [[0, Y], [X, X+1]])


def test_det():
    """Test for numpoly.det."""
    array = [[1, 2], [3, 4]]
    poly = polynomial([[1, Y], [X, 1]])
    assert_equal(numpoly.det([array, poly]), [-2, 1-X*Y])
    assert_equal(numpy.linalg.det(poly), 1-X*Y)
    assert_equal(
        numpoly.det([[1, X, Y], [Y, 1, X], [X, Y, 1]]), X**3+Y**3-3*X*Y+1)


def test_diag(func_interface):
    """Tests for numpoly.diag."""
    poly = polynomial([[1, 2, X], [4, Y, 6], [7, 8, X+Y]])
    assert_equal(func_interface.diag(poly), [1, Y, X+Y])
    assert_equal(func_interface.diag(poly, k=1), [2, 6])
    assert_equal(func_interface.diag(poly, k=-1), [4, 8])

    poly = polynomial([X, Y])
    assert_equal(func_interface.diag(poly), [[X, 0], [0, Y]])
    assert_equal(func_interface.diag(poly, k=1),
                 [[0, X, 0], [0, 0, Y], [0, 0, 0]])
    assert_equal(func_interface.diag(poly, k=-1),
                 [[0, 0, 0], [X, 0, 0], [0, Y, 0]])


def test_diagonal(interface):
    """Tests for numpoly.diagonal."""
    # TODO: return view instead of copy
    poly = polynomial([[1, 2, X], [4, Y, 6], [7, 8, X+Y]])
    assert_equal(interface.diagonal(poly), [1, Y, X+Y])
    assert_equal(interface.diagonal(poly, offset=1), [2, 6])
    assert_equal(interface.diagonal(poly, offset=-1), [4, 8])

    poly = numpoly.monomial(27).reshape(3, 3, 3)
    assert_equal(interface.diagonal(poly, axis1=0, axis2=1),
                 [[1, X**12, X**24],
                  [X, X**13, X**25],
                  [X**2, X**14, X**26]])
    assert_equal(interface.diagonal(poly, axis1=0, axis2=2),
                 [[1, X**10, X**20],
                  [X**3, X**13, X**23],
                  [X**6, X**16, X**26]])
    assert_equal(interface.diagonal(poly, axis1=1, axis2=2),
                 [[1, X**4, X**8],
                  [X**9, X**13, X**17],
                  [X**18, X**22, X**26]])


def test_diff(func_interface):
    """Tests for numpoly.diff."""
    poly = polynomial([[1, 2, X], [4, Y, 6], [7, 8, X+Y]])
    assert_equal(func_interface.diff(poly),
                 [[1, X-2], [Y-4, 6-Y], [1, X+Y-8]])
    assert_equal(func_interface.diff(poly, n=2),
                 [[X-3], [10-2*Y], [X+Y-9]])
    assert_equal(func_interface.diff(poly, axis=0),
                 [[3, Y-2, 6-X], [3, 8-Y, X+Y-6]])
    assert_equal(func_interface.diff(poly, append=X),
                 [[1, X-2, 0], [Y-4, 6-Y, X-6], [1, X+Y-8, -Y]])
    assert_equal(func_interface.diff(poly, prepend=Y),
                 [[1-Y, 1, X-2], [4-Y, Y-4, 6-Y], [7-Y, 1, X+Y-8]])
    assert_equal(func_interface.diff(poly, append=X, prepend=Y),
                 [[1-Y, 1, X-2, 0], [4-Y, Y-4, 6-Y, X-6], [7-Y, 1, X+Y-8, -Y]])


def test_divmod(func_interface):
    """Tests for numpoly.divmod."""
    array = numpy.array([7, 11])
    quotient, remainder = func_interface.divmod(array, 5)
    assert_equal(quotient, [1, 2])
    assert_equal(remainder, [2, 1])
    with raises(numpoly.FeatureNotSupported):
        func_interface.divmod(array, X)
    with raises(numpoly.FeatureNotSupported):
        func_interface.divmod(X, X)


def test_dsplit(func_interface):
    """Tests for numpoly.dsplit."""
    poly = numpoly.polynomial([[[1, X], [X+Y, Y]]])
    part1, part2 = func_interface.dsplit(poly, 2)
    assert_equal(part1, [[[1], [X+Y]]])
    assert_equal(part2, [[[X], [Y]]])


def test_dstack(func_interface):
    """Tests for numpoly.dstack."""
    poly1 = numpoly.polynomial([1, X, 2])
    poly2 = numpoly.polynomial([Y, 3, 4])
    assert_equal(func_interface.dstack([poly1, poly2]),
                 [[[1, Y], [X, 3], [2, 4]]])


def test_ediff1d(func_interface):
    """Tests for numpoly.ediff1d."""
    poly1 = numpoly.polynomial([1, X, 2])
    assert_equal(func_interface.ediff1d(poly1), [X-1, 2-X])
    poly2 = numpoly.polynomial([Y, 3, 4])
    assert_equal(func_interface.ediff1d(poly2), [3-Y, 1])


def test_expand_dims(func_interface):
    """Tests for numpoly.expand_dims."""
    poly1 = numpoly.polynomial([[1, X], [Y, 2], [3, 4]])
    assert func_interface.expand_dims(poly1, axis=0).shape == (1, 3, 2)
    assert func_interface.expand_dims(poly1, axis=1).shape == (3, 1, 2)
    assert func_interface.expand_dims(poly1, axis=2).shape == (3, 2, 1)
    array = numpy.arange(12).reshape(2, 3, 2)
    assert func_interface.expand_dims(array, axis=1).shape == (2, 1, 3, 2)


def test_equal(interface):
    """Tests for numpoly.equal."""
    poly = polynomial([[0, 2+Y], [X, 2]])
    assert_equal(interface.equal(X, X), True)
    assert_equal(interface.equal(X, [X]), [True])
    assert_equal([X] == X, [True])
    assert_equal(interface.equal(X, Y), False)
    assert_equal(([X, 2+Y] == poly), [[False, True], [True, False]])
    assert_equal(interface.equal(numpoly.polynomial([X, 2+Y]), poly),
                 [[False, True], [True, False]])
    assert_equal((X == poly), [[False, False], [True, False]])
    assert_equal(interface.equal(X, poly), [[False, False], [True, False]])
    assert_equal(poly == poly.T, [[True, False], [False, True]])
    assert_equal(interface.equal(poly, poly.T), [[True, False], [False, True]])


def test_floor(func_interface):
    """Tests for numpoly.floor."""
    poly = polynomial([-1.7*X, X-1.5, -0.2, 3.2+1.5*X, 1.7, 2.0])
    assert_equal(func_interface.floor(poly),
                 [-2.0*X, -2.0+X, -1.0, 3.0+X, 1.0, 2.0])


def test_floor_divide(interface):
    """Tests for numpoly.floor_divide."""
    poly = polynomial([[0., 2.*Y], [X, 2.]])
    assert_equal(interface.floor_divide(poly, 2), [[0., Y], [0., 1]])
    assert_equal(interface.floor_divide(poly, [1., 2.]), [[0., Y], [X, 1]])
    assert_equal(interface.floor_divide(
        poly, [[1., 2.], [2., 1.]]), [[0., Y], [0., 2.]])
    with raises(numpoly.FeatureNotSupported):
        interface.floor_divide(poly, X)
    with raises(numpoly.FeatureNotSupported):
        poly.__floordiv__(poly)
    with raises(numpoly.FeatureNotSupported):
        poly.__rfloordiv__(poly)

    out = numpoly.ndpoly(
        exponents=poly.exponents,
        shape=(2, 2),
        names=("q0", "q1"),
        dtype=float,
    )
    numpoly.floor_divide(poly, 2, out=out)
    assert_equal(out, [[0., Y], [0., 1.]])


def test_full(func_interface):
    """Tests for numpoly.full."""
    assert_equal(numpoly.full((3,), X), [X, X, X])
    assert_equal(numpoly.aspolynomial(
        func_interface.full((3,), 1.*X)), [1.*X, X, X])
    assert_equal(numpoly.full((3,), Y, dtype=float), [1.*Y, Y, Y])
    if func_interface is numpy:  # fails in numpy, but only with func dispatch.
        with raises(ValueError, ):
            assert_equal(numpy.full((3,), Y, dtype=float), [1.*Y, Y, Y])
            raise ValueError


def test_full_like(func_interface):
    """Tests for numpoly.full_like."""
    poly = numpoly.polynomial([1, X, 2])
    assert_equal(func_interface.full_like(poly, X), [X, X, X])
    assert_equal(numpoly.full_like([1, X, 2], X), [X, X, X])
    poly = numpoly.polynomial([1., X, 2])
    assert_equal(func_interface.full_like(poly, Y), [1.*Y, Y, Y])


def test_greater(interface):
    """Tests for numpoly.greater."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(interface.greater(poly, X),
                 [[False, False, False, True],
                  [True, True, True, False],
                  [False, True, False, False],
                  [True, True, True, True]])
    assert_equal(interface.greater(poly, Y),
                 [[False, False, False, True],
                  [False, False, True, False],
                  [False, True, False, False],
                  [True, True, False, False]])


def test_greater_equal(interface):
    """Tests for numpoly.greater_equal."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(interface.greater_equal(poly, X),
                 [[False, True, False, True],
                  [True, True, True, False],
                  [False, True, False, True],
                  [True, True, True, True]])
    assert_equal(interface.greater_equal(poly, Y),
                 [[False, False, False, True],
                  [True, False, True, False],
                  [False, True, False, False],
                  [True, True, True, False]])


def test_hsplit(func_interface):
    """Tests for numpoly.hsplit."""
    poly = numpoly.polynomial([[1, X, X**2], [X+Y, Y, Y]])
    part1, part2, part3 = func_interface.hsplit(poly, 3)
    assert_equal(part1, [[1], [X+Y]])
    assert_equal(part2, [[X], [Y]])
    assert_equal(part3, [[X**2], [Y]])


def test_hstack(func_interface):
    """Tests for numpoly.hstack."""
    poly1 = numpoly.polynomial([1, X, 2])
    poly2 = numpoly.polynomial([Y, 3, 4])
    assert_equal(func_interface.hstack([poly1, poly2]),
                 [1, X, 2, Y, 3, 4])


def test_less(interface):
    """Tests for numpoly.less."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(interface.less(poly, X),
                 [[True, False, True, False],
                  [False, False, False, True],
                  [True, False, True, False],
                  [False, False, False, False]])
    assert_equal(interface.less(poly, Y),
                 [[True, True, True, False],
                  [False, True, False, True],
                  [True, False, True, True],
                  [False, False, False, True]])


def test_less_equal(interface):
    """Tests for numpoly.less_equal."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(interface.less_equal(poly, X),
                 [[True, True, True, False],
                  [False, False, False, True],
                  [True, False, True, True],
                  [False, False, False, False]])
    assert_equal(interface.less_equal(poly, Y),
                 [[True, True, True, False],
                  [True, True, False, True],
                  [True, False, True, True],
                  [False, False, True, True]])


def test_inner(func_interface):
    """Tests for numpoly.inner."""
    poly1, poly2 = polynomial([[0, Y], [X+1, 1]])
    assert_equal(func_interface.inner(poly1, poly2), Y)


def test_isclose(func_interface):
    """Tests for numpoly.isclose."""
    poly1 = numpoly.polynomial([1e10*X, 1e-7])
    poly2 = numpoly.polynomial([1.00001e10*X, 1e-8])
    assert_equal(func_interface.isclose(poly1, poly2), [True, False])
    poly1 = numpoly.polynomial([1e10*X, 1e-8])
    poly2 = numpoly.polynomial([1.00001e10*X, 1e-9])
    assert_equal(func_interface.isclose(poly1, poly2), [True, True])
    poly2 = numpoly.polynomial([1e10*Y, 1e-8])
    assert_equal(func_interface.isclose(poly1, poly2), [False, True])


def test_isfinite(func_interface):
    """Tests for numpoly.isfinite."""
    assert_equal(func_interface.isfinite(X), True)
    assert_equal(func_interface.isfinite(numpy.nan*X), False)
    poly = numpoly.polynomial([numpy.log(-1.), X, numpy.log(0)])
    assert_equal(func_interface.isfinite(poly), [False, True, False])
    out = numpy.ones(3, dtype=bool)
    print(out)
    func_interface.isfinite(poly, out=(out,))
    print(out)
    assert_equal(out, [False, True, False])


def test_logical_and(func_interface):
    """Tests for numpoly.logical_and."""
    poly1 = numpoly.polynomial([0, X])
    poly2 = numpoly.polynomial([1, X])
    poly3 = numpoly.polynomial([0, Y])
    assert_equal(func_interface.logical_and(1, poly1), [False, True])
    assert_equal(1 and poly1, poly1)
    assert_equal(func_interface.logical_and(1, poly2), [True, True])
    assert_equal(1 and poly2, poly2)
    assert_equal(func_interface.logical_and(poly2, poly3), [False, True])


def test_logical_or(func_interface):
    """Tests for numpoly.logical_or."""
    poly1 = numpoly.polynomial([0, X])
    poly2 = numpoly.polynomial([1, X])
    poly3 = numpoly.polynomial([0, Y])
    assert_equal(func_interface.logical_or(1, poly1), [True, True])
    assert_equal(1 or poly1, 1, type_=int)
    assert_equal(func_interface.logical_or(0, poly1), [False, True])
    assert_equal(0 or poly1, poly1)
    assert_equal(func_interface.logical_or(poly2, poly3), [True, True])


def test_matmul(func_interface):
    """Tests for numpoly.matmul."""
    poly1 = numpoly.polynomial([[0, X], [1, Y]])
    poly2 = numpoly.polynomial([X, 2])
    assert_equal(func_interface.matmul(
        poly1, poly2), [[X**2, 2*X], [X+X*Y, 2+2*Y]])
    assert_equal(func_interface.matmul(
        numpy.ones((2, 5, 6, 4)), numpy.ones((2, 5, 4, 3))),
                 4*numpy.ones((2, 5, 6, 3)))
    with raises(ValueError):
        func_interface.matmul(poly1, 4)
    with raises(ValueError):
        func_interface.matmul(3, poly2)


def test_mean(interface):
    """Tests for numpoly.mean."""
    poly = numpoly.polynomial([[1, 2*X], [3*Y+X, 4]])
    assert_equal(interface.mean(poly), 1.25+0.75*Y+0.75*X)
    assert_equal(interface.mean(poly, axis=0), [0.5+1.5*Y+0.5*X, 2.0+X])
    assert_equal(interface.mean(poly, axis=1), [0.5+X, 2.0+1.5*Y+0.5*X])


def test_max(interface):
    """Tests for numpoly.max."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(interface.max(poly), X**4)
    assert_equal(interface.max(poly, axis=0), [X**4, Y**2, X**2, Y**2])
    assert_equal(interface.max(poly, axis=1), [X**2, X**2, Y**2, X**4])
    assert_equal(interface.max(poly.reshape(2, 2, 2, 2), axis=(0, 1)),
                 [[X**4, Y**2], [X**2, Y**2]])


def test_maximum(func_interface):
    """Tests for numpoly.maximum."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(func_interface.maximum(poly, X),
                 [[X, X, X, X**2], [Y, Y-1, Y**2, X],
                  [X, X**2, X, X], [Y**2, X**4, Y, Y-1]])
    assert_equal(func_interface.maximum(poly, Y),
                 [[Y, Y, Y, X**2], [Y, Y, Y**2, Y],
                  [Y, X**2, Y, Y], [Y**2, X**4, Y, Y]])


def test_min(interface):
    """Tests for numpoly.min."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(interface.min(poly), 1)
    assert_equal(interface.min(poly, axis=0), [1, 3, 2, X])
    assert_equal(interface.min(poly, axis=1), [1, 2, 3, Y])
    assert_equal(interface.min(poly.reshape(2, 2, 2, 2), axis=(0, 1)),
                 [[1, 3], [2, X]])


def test_minimum(func_interface):
    """Tests for numpoly.minimum."""
    poly = numpoly.polynomial([[1, X, X-1, X**2],
                               [Y, Y-1, Y**2, 2],
                               [X-1, X**2, 3, X],
                               [Y**2, X**4, Y, Y-1]])
    assert_equal(func_interface.minimum(poly, X),
                 [[1, X, X-1, X], [X, X, X, 2],
                  [X-1, X, 3, X], [X, X, X, X]])
    assert_equal(func_interface.minimum(poly, Y),
                 [[1, X, X-1, Y], [Y, Y-1, Y, 2],
                  [X-1, Y, 3, X], [Y, Y, Y, Y-1]])


def test_moveaxis():
    """Tests for numpoly.moveaxis."""
    # np.moveaxis dispatching doesn't seem to work
    x = numpy.arange(6).reshape(1, 2, 3)
    assert_equal(numpoly.moveaxis(x, 0, -1),
                 [[[0], [1], [2]], [[3], [4], [5]]])
    assert_equal(numpoly.moveaxis(x, [0, 2], [2, 0]),
                 [[[0], [3]], [[1], [4]], [[2], [5]]])


def test_multiply(func_interface):
    """Tests for numpoly.multiply."""
    poly = polynomial([[0, 2+Y], [X, 2]])
    assert_equal(2*poly, [[0, 4+2*Y], [2*X, 4]])
    assert_equal(func_interface.multiply(2, poly), [[0, 4+2*Y], [2*X, 4]])
    assert_equal([X, 1]*poly, [[0, 2+Y], [X*X, 2]])
    assert_equal(func_interface.multiply([X, 1], poly), [[0, 2+Y], [X*X, 2]])
    assert_equal([[X, 1], [Y, 0]]*poly, [[0, 2+Y], [X*Y, 0]])
    assert_equal(
        func_interface.multiply([[X, 1], [Y, 0]], poly), [[0, 2+Y], [X*Y, 0]])


def test_negative(func_interface):
    """Tests for numpoly.negative."""
    poly = polynomial([[X, -Y], [-4, Y]])
    assert_equal(-(X-Y-1), 1-X+Y)
    assert_equal(func_interface.negative(X-Y-1), 1-X+Y)
    assert_equal(func_interface.negative(poly), [[-X, Y], [4, -Y]])


def test_nonzero(interface):
    """Tests for numpoly.nonzero."""
    poly = polynomial([[3*X, 0, 0], [0, 4*Y, 0], [5*X+Y, 6*X, 0]])
    assert_equal(poly[interface.nonzero(poly)], [3*X, 4*Y, 5*X+Y, 6*X])


def test_not_equal(interface):
    """Tests for numpoly.not_equal."""
    poly = polynomial([[0, 2+Y], [X, 2]])
    assert_equal(([X, 2+Y] != poly), [[True, False], [False, True]])
    assert_equal(interface.not_equal(numpoly.polynomial([X, 2+Y]), poly),
                 [[True, False], [False, True]])
    assert_equal((X != poly), [[True, True], [False, True]])
    assert_equal(interface.not_equal(X, poly), [[True, True], [False, True]])
    assert_equal(poly != poly.T, [[False, True], [True, False]])
    assert_equal(
        interface.not_equal(poly, poly.T), [[False, True], [True, False]])


def test_ones_like(func_interface):
    """Tests for numpoly.ones_like."""
    poly = numpoly.polynomial([1, X, 2])
    assert_equal(func_interface.ones_like(poly), [1, 1, 1])
    assert_equal(numpoly.ones_like([1, X, 2]), [1, 1, 1])
    poly = numpoly.polynomial([1., X, 2])
    assert_equal(func_interface.ones_like(poly), [1., 1., 1.])


def test_outer(func_interface):
    """Tests for numpoly.outer."""
    poly1, poly2 = polynomial([[0, Y], [X+1, 1]])
    assert_equal(func_interface.outer(poly1, poly2), [[0, 0], [X*Y+Y, Y]])


def test_positive(func_interface):
    """Tests for numpoly.positive."""
    poly = polynomial([[0, Y], [X, 1]])
    assert_equal(poly, +poly)
    assert_equal(poly, func_interface.positive(poly))
    assert poly is not +poly
    assert poly is not func_interface.positive(poly)


def test_power(func_interface):
    """Tests for numpoly.power."""
    poly = polynomial([[0, Y], [X-1, 2]])
    assert_equal(X**[2], [X**2])
    assert_equal(func_interface.power(X, [2]), [X**2])
    assert_equal(polynomial([X])**[2], [X**2])
    assert_equal(func_interface.power(polynomial([X]), [2]), [X**2])
    assert_equal(polynomial([X, Y])**[2], [X**2, Y**2])
    assert_equal(func_interface.power(polynomial([X, Y]), [2]), [X**2, Y**2])
    assert_equal(polynomial([X])**[1, 2], [X, X**2])
    assert_equal(func_interface.power(polynomial([X]), [1, 2]), [X, X**2])
    assert_equal((X*Y)**[0, 1, 2, 3], [1, X*Y, X**2*Y**2, X**3*Y**3])
    assert_equal(func_interface.power(
        X*Y, [0, 1, 2, 3]), [1, X*Y, X**2*Y**2, X**3*Y**3])
    assert_equal(poly ** 2, [[0, Y**2], [X*X-2*X+1, 4]])
    assert_equal(func_interface.power(poly, 2), [[0, Y**2], [X*X-2*X+1, 4]])
    assert_equal(poly ** [1, 2], [[0, Y**2], [X-1, 4]])
    assert_equal(func_interface.power(poly, [1, 2]), [[0, Y**2], [X-1, 4]])
    assert_equal(poly ** [[1, 2], [2, 1]], [[0, Y**2], [X*X-2*X+1, 2]])
    assert_equal(func_interface.power(
        poly, [[1, 2], [2, 1]]), [[0, Y**2], [X*X-2*X+1, 2]])


def test_prod(interface):
    """Tests for numpoly.prod."""
    poly = numpoly.polynomial([[1, X, X**2], [X+Y, Y, Y]])
    assert_equal(interface.prod(poly), X**3*Y**3+X**4*Y**2)
    assert_equal(interface.prod(poly, axis=0), [Y+X, X*Y, X**2*Y])


def test_remainder(func_interface):
    """Tests for numpoly.remainder."""
    assert_equal(func_interface.remainder([7, 11], 5), [2, 1])
    with raises(numpoly.FeatureNotSupported):
        func_interface.remainder(X, X)
    with raises(numpoly.FeatureNotSupported):
        func_interface.remainder([1, 2], X)


def test_repeat(func_interface):
    """Tests for numpoly.repeat."""
    poly = numpoly.polynomial([[1, X-1], [X**2, X]])
    assert_equal(func_interface.repeat(poly, 2),
                 [[1, -1+X], [1, -1+X], [X**2, X], [X**2, X]])
    assert_equal(func_interface.repeat(poly, 3, axis=1),
                 [[1, 1, 1, -1+X, -1+X, -1+X],
                  [X**2, X**2, X**2, X, X, X]])
    assert_equal(numpoly.repeat(poly, [1, 2], axis=0),
                 [[1, -1+X], [X**2, X], [X**2, X]])


def test_reshape(interface):
    """Tests for numpoly.reshape."""
    poly = numpoly.polynomial([[1, X, X**2], [X+Y, Y, Y]])
    assert_equal(interface.reshape(poly, (3, 2)),
                 [[1, X], [X**2, X+Y], [Y, Y]])
    assert_equal(interface.reshape(poly, 6),
                 [1, X, X**2, X+Y, Y, Y])


def test_result_type(func_interface):
    """Tests for numpoly.result_type."""
    dtypes = ["uint8", "int16", "float32", "complex64"]
    dtypes = [numpy.dtype(dtype) for dtype in dtypes]
    for idx, dtype1 in enumerate(dtypes):
        for dtype2 in dtypes[idx:]:
            assert func_interface.result_type(3, dtype1, dtype2) == dtype2
            assert func_interface.result_type(
                numpoly.variable(dtype=dtype1),
                numpy.arange(3, dtype=dtype2),
            ) == dtype2


def test_rint(func_interface):
    """Tests for numpoly.rint."""
    poly = numpoly.polynomial([-1.7*X, X-1.5])
    assert_equal(func_interface.rint(poly), [-2.*X, X-2.])


def test_roots(func_interface):
    func_interface = numpoly
    poly = [3.2, 2, 1]
    assert_equal(func_interface.roots(poly),
                 [-0.3125+0.46351241j, -0.3125-0.46351241j])
    poly = 3.2*Y**2+2*Y+1
    assert_equal(func_interface.roots(poly),
                 [-0.3125+0.46351241j, -0.3125-0.46351241j])
    with raises(ValueError):
        func_interface.roots(X*Y)
    with raises(ValueError):
        func_interface.roots([[X, 1], [2, X]])


def test_split(func_interface):
    """Tests for numpoly.split."""
    poly = numpoly.polynomial([[1, X, X**2], [X+Y, Y, Y]])
    part1, part2 = func_interface.split(poly, 2, axis=0)
    assert_equal(part1, [[1, X, X**2]])
    assert_equal(part2, [[X+Y, Y, Y]])
    part1, part2, part3 = func_interface.split(poly, 3, axis=1)
    assert_equal(part1, [[1], [X+Y]])
    assert_equal(part2, [[X], [Y]])
    assert_equal(part3, [[X**2], [Y]])


def test_square(func_interface):
    """Tests for numpoly.square."""
    assert_equal(func_interface.square(X+Y), X**2+2*X*Y+Y**2)
    assert_equal((1+X)**2, 1+2*X+X**2)


def test_stack(func_interface):
    """Tests for numpoly.stack."""
    poly = polynomial([1, X, Y])
    assert_equal(
        func_interface.stack([poly, poly], axis=0), [[1, X, Y], [1, X, Y]])
    assert_equal(
        func_interface.stack([poly, poly], axis=1), [[1, 1], [X, X], [Y, Y]])


def test_subtract(func_interface):
    """Tests for numpoly.subtract."""
    assert_equal(-X+3, 3-X)
    assert_equal(4 - polynomial([1, X, Y]), [3, 4-X, 4-Y])
    assert_equal(
        func_interface.subtract(4, polynomial([1, X, Y])), [3, 4-X, 4-Y])
    assert_equal(polynomial([0, X]) - polynomial([Y, 0]), [-Y, X])
    assert_equal(func_interface.subtract(polynomial([0, X]), [Y, 0]), [-Y, X])
    assert_equal(polynomial([[1, X], [Y, X*Y]]) - [2, X],
                 [[-1, 0], [Y-2, X*Y-X]])
    assert_equal(
        func_interface.subtract(polynomial([[1, X], [Y, X*Y]]), [2, X]),
        [[-1, 0], [Y-2, X*Y-X]])


def test_sum(interface):
    """Tests for numpoly.sum."""
    poly = polynomial([[1, 5*X], [X+3, -Y]])
    assert_equal(interface.sum(poly), -Y+X*6+4)
    assert_equal(interface.sum(poly, axis=0), [X+4, -Y+X*5])
    assert_equal(
        interface.sum(poly, axis=-1, keepdims=True), [[X*5+1], [X-Y+3]])


def test_transpose(func_interface):
    """Tests for numpoly.transpose."""
    poly = numpoly.polynomial([[1, X-1], [X**2, X]])
    assert_equal(func_interface.transpose(poly),
                 [[1, X**2], [X-1, X]])
    assert_equal(
        poly.T, [[1, X**2], [X-1, X]], c_contiguous=False, f_contiguous=True)


def test_true_divide(func_interface):
    """Tests for numpoly.true_divide."""
    poly = polynomial([[0, Y], [X, 1]])
    assert_equal(func_interface.true_divide(
        poly, 2), polynomial([[0, 0.5*Y], [0.5*X, 0.5]]))
    assert_equal(func_interface.true_divide(
        poly, [1, 2]), [[0, 0.5*Y], [X, 0.5]])
    assert_equal(func_interface.true_divide(
        poly, [[1, 2], [2, 1]]), [[0, 0.5*Y], [0.5*X, 1]])
    with raises(numpoly.FeatureNotSupported):
        func_interface.true_divide(poly, X)


def test_vsplit(func_interface):
    """Tests for numpoly.vsplit."""
    poly = numpoly.polynomial([[1, X, X**2], [X+Y, Y, Y]])
    part1, part2 = func_interface.vsplit(poly, 2)
    assert_equal(part1, [[1, X, X**2]])
    assert_equal(part2, [[X+Y, Y, Y]])


def test_vstack(func_interface):
    """Tests for numpoly.vstack."""
    poly1 = numpoly.polynomial([1, X, 2])
    poly2 = numpoly.polynomial([Y, 3, 4])
    assert_equal(func_interface.vstack([poly1, poly2]), [[1, X, 2], [Y, 3, 4]])


def test_where(func_interface):
    """Tests for numpoly.where."""
    poly1 = numpoly.polynomial([0, 4, 0])
    poly2 = numpoly.polynomial([Y, 0, X])
    assert_equal(func_interface.where([0, 1, 0], poly1, poly2), [Y, 4, X])
    assert_equal(func_interface.where([1, 0, 1], poly1, poly2), [0, 0, 0])
    assert_equal(func_interface.where(poly1)[0], [1])
    assert_equal(func_interface.where(poly2)[0], [0, 2])


def test_zeros_like(func_interface):
    """Tests for numpoly.zeros_like."""
    poly = numpoly.polynomial([1, X, 2])
    assert_equal(func_interface.zeros_like(poly), [0, 0, 0])
    assert_equal(numpoly.zeros_like([1, X, 2]), [0, 0, 0])
    poly = numpoly.polynomial([1., X, 2])
    assert_equal(func_interface.zeros_like(poly), [0., 0., 0.])



"""Evaluate polynomial by inserting new values in to the indeterminants."""
from __future__ import annotations
from typing import Dict, Sequence, Optional, Union
import logging

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike


def call(
    poly: PolyLike,
    args: Sequence[Optional[PolyLike]] = (),
    kwargs: Dict[str, PolyLike] = None,
) -> Union[numpy.ndarray, ndpoly]:
    """
    Evaluate polynomial by inserting new values in to the indeterminants.

    Equivalent to calling the polynomial or using the ``__call__`` method.

    Args:
        poly:
            Polynomial to evaluate.
        args:
            Argument to evaluate indeterminants. Ordered positional by
            ``poly.indeterminants``. None values indicate that a variable is
            not to be evaluated, creating a partial evaluation.
        kwargs:
            Same as ``args``, but positioned by name.

    Return:
        Evaluated polynomial. If the resulting array does not contain any
        indeterminants, an array is returned instead of a polynomial.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[q0, q0-1], [q1, q1+q0]])
        >>> numpoly.call(poly)
        polynomial([[q0, q0-1],
                    [q1, q1+q0]])
        >>> poly
        polynomial([[q0, q0-1],
                    [q1, q1+q0]])
        >>> numpoly.call(poly, (1, 0))
        array([[1, 0],
               [0, 1]])
        >>> numpoly.call(poly, (1,), {"q1": [0, 1, 2]})
        array([[[1, 1, 1],
                [0, 0, 0]],
        <BLANKLINE>
               [[0, 1, 2],
                [1, 2, 3]]])
        >>> numpoly.call(poly, (q1,))
        polynomial([[q1, q1-1],
                    [q1, 2*q1]])
        >>> numpoly.call(poly, kwargs={"q1": q0-1, "q0": 2*q1})
        polynomial([[2*q1, 2*q1-1],
                    [q0-1, 2*q1+q0-1]])

    """
    logger = logging.getLogger(__name__)

    poly = numpoly.aspolynomial(poly)
    kwargs = kwargs if kwargs else {}

    parameters = dict(zip(poly.names, poly.indeterminants))
    if kwargs:
        parameters.update(kwargs)
    for arg, name in zip(args, poly.names):
        if name in kwargs:
            raise TypeError(f"multiple values for argument '{name}'")
        if arg is not None:
            parameters[name] = arg
    extra_args = [key for key in parameters if key not in poly.names]
    if extra_args:
        raise TypeError(f"unexpected keyword argument '{extra_args[0]}'")

    # There can only be one shape:
    ones = numpy.ones((), dtype=int)
    for value in parameters.values():
        ones = ones * numpy.ones(numpoly.polynomial(value).shape, dtype=int)
    shape = poly.shape + ones.shape

    logger.debug("poly shape: %s", poly.shape)
    logger.debug("parameter common shape: %s", ones.shape)
    logger.debug("output shape: %s", shape)
    # main loop:
    out = numpy.zeros((), dtype=int)
    for exponent, coefficient in zip(poly.exponents, poly.coefficients):
        term = ones
        for power, name in zip(exponent, poly.names):
            term = term * parameters[name] ** power
        if isinstance(term, numpoly.ndpoly):
            tmp = numpoly.outer(coefficient, term)
        else:
            tmp = numpy.outer(coefficient, term)
        out = out + tmp.reshape(shape)

    if isinstance(out, numpoly.ndpoly):
        if out.isconstant():
            out = out.tonumpy()
        else:
            out, _ = numpoly.align_indeterminants(out, poly.indeterminants)

    return out



"""Gives a new shape to an array without changing its data."""
from __future__ import annotations
from typing import Any, Sequence

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements

Order = Any
try:
    from typing import Literal, Union

    Order = Union[None, Literal["C"], Literal["F"], Literal["A"]]  # type: ignore
except ImportError:
    pass


@implements(numpy.reshape)
def reshape(
    a: PolyLike,
    /,
    shape: Union[int, Sequence[int]] = None,
    order: Order = "C",
    newshape: Union[int, Sequence[int]] = None,
) -> ndpoly:
    """
    Give a new shape to an array without changing its data.

    Args:
        a:
            Array to be reshaped.
        shape:
            The new shape should be compatible with the original shape. If an
            integer, then the result will be a 1-D array of that length. One
            shape dimension can be -1. In this case, the value is inferred from
            the length of the array and remaining dimensions.
        order:
            Read the elements of `a` using this index order, and place the
            elements into the reshaped array using this index order.  'C' means
            to read / write the elements using C-like index order, with the
            last axis index changing fastest, back to the first axis index
            changing slowest. 'F' means to read / write the elements using
            Fortran-like index order, with the first index changing fastest,
            and the last index changing slowest. Note that the 'C' and 'F'
            options take no account of the memory layout of the underlying
            array, and only refer to the order of indexing. 'A' means to read
            / write the elements in Fortran-like index order if `a` is Fortran
            *contiguous* in memory, C-like order otherwise.

    Return:
        This will be a new view object if possible; otherwise, it will be a
        copy.  Note there is no guarantee of the *memory layout* (C- or
        Fortran- contiguous) of the returned array.

    Example:
        >>> numpoly.reshape([1, 2, 3, 4], (2, 2))
        polynomial([[1, 2],
                    [3, 4]])
        >>> numpoly.reshape(numpoly.monomial(6), (3, 2))
        polynomial([[1, q0],
                    [q0**2, q0**3],
                    [q0**4, q0**5]])

    """
    poly = numpoly.aspolynomial(a)
    array = numpy.reshape(poly.values, shape=shape, newshape=newshape, order=order)
    return numpoly.aspolynomial(array, names=poly.indeterminants)



"""Compute the determinant of an polynomial array."""
from __future__ import annotations

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.linalg.det)
def det(a: PolyLike) -> ndpoly:
    """
    Compute the determinant of an polynomial array.

    Args:
        a:
            Input array to compute determinants for.
            Shape on form `(..., M, M)`.

    Return:
        Determinant of `a`. Shape `(...)`.

    Note:
        Broadcasting rules apply, see the `numpy.linalg` documentation for
        details.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> a = numpoly.polynomial([[[1, 2], [3, 4]], [[1, q0], [q1, 1]]])
        >>> numpoly.det(a)
        polynomial([-2, -q0*q1+1])

    """
    a = numpoly.aspolynomial(a)
    assert a.ndim >= 2, a
    assert a.shape[-2] == a.shape[-1], a.shape
    dims = a.shape[-1]
    index = (slice(None),) * (a.ndim - 2)
    if dims == 2:
        return (
            a[index + (0, 0)] * a[index + (1, 1)]
            - a[index + (1, 0)] * a[index + (0, 1)]
        )
    out = numpoly.zeros_like(a, shape=a.shape[:-2])
    r = numpy.arange(1, dims, dtype=int)
    for idx in range(dims):
        idx0 = index + (0, idx)
        idx1 = index + (slice(1, None), (r + idx) % dims)
        out = out + a[idx0] * det(a[idx1])
    return out



"""Apply a function repeatedly over multiple axes."""
from __future__ import annotations
from functools import wraps
from typing import Callable

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.apply_over_axes)
def apply_over_axes(
    func: Callable,
    a: PolyLike,
    axes: numpy.typing.ArrayLike,
) -> ndpoly:
    """
    Apply a function repeatedly over multiple axes.

    `func` is called as `res = func(a, axis)`, where `axis` is the first
    element of `axes`.  The result `res` of the function call must have either
    the same dimensions as `a` or one less dimension.  If `res` has one less
    dimension than `a`, a dimension is inserted before `axis`.  The call to
    `func` is then repeated for each axis in `axes`, with `res` as the first
    argument.

    Args:
        func:
            This function must take two arguments, `func(a, axis)`.
        a:
            Input array.
        axes:
            Axes over which `func` is applied; the elements must be integers.

    Return:
        The output array.  The number of dimensions is the same as `a`, but
        the shape can be different.  This depends on whether `func` changes
        the shape of its output with respect to its input.

    Example:
        >>> polynomial = numpy.arange(24).reshape(2, 4, 3)
        >>> polynomial = polynomial*numpoly.variable(3)
        >>> numpoly.apply_over_axes(
        ...     func=numpoly.sum, a=polynomial, axes=1)
        polynomial([[[18*q0, 22*q1, 26*q2]],
        <BLANKLINE>
                    [[66*q0, 70*q1, 74*q2]]])
        >>> numpoly.apply_over_axes(
        ...     func=numpoly.sum, a=polynomial, axes=[0, 2])
        polynomial([[[16*q2+14*q1+12*q0],
                     [22*q2+20*q1+18*q0],
                     [28*q2+26*q1+24*q0],
                     [34*q2+32*q1+30*q0]]])

    """

    @wraps(func)
    def wrapper_func(array, axis):
        """Wrap func function."""
        # Align indeterminants in case slicing changed them
        array = numpoly.polynomial(array, names=a.indeterminants)
        array, _ = numpoly.align.align_indeterminants(array, a.indeterminants)
        # Evaluate function
        out = func(array, axis=axis)
        # Restore indeterminants in case func changed them.
        out, _ = numpoly.align.align_indeterminants(out, a.indeterminants)
        return out

    # Initiate wrapper
    a = numpoly.aspolynomial(a)
    out = numpy.apply_over_axes(wrapper_func, a=a.values, axes=axes)
    return out



import pytest
import numpy
import numpoly

POLYNOMIAL_CONFIGURATIONS = [
    dict(zip(["display_graded", "display_inverse", "display_reverse", "expected_output"], args)
         ) for args in [
             (False, False, False, "1+q0+q0**2+q1+q0*q1+q1**2"),
             (False, True, False, "q1**2+q0*q1+q1+q0**2+q0+1"),
             (False, False, True, "1+q1+q1**2+q0+q0*q1+q0**2"),
             (False, True, True, "q0**2+q0*q1+q0+q1**2+q1+1"),
             (True, False, False, "1+q0+q1+q0**2+q0*q1+q1**2"),
             (True, True, False, "q1**2+q0*q1+q0**2+q1+q0+1"),
             (True, False, True, "1+q1+q0+q1**2+q0*q1+q0**2"),
             (True, True, True, "q0**2+q0*q1+q1**2+q0+q1+1"),
]]


@pytest.fixture(params=POLYNOMIAL_CONFIGURATIONS)
def display_config(request):
    """Parameterization of various display options."""
    yield request.param


def test_display_order(display_config):
    """Ensure string output changes with various display options."""
    expected_output = display_config.pop("expected_output")
    polynomial = numpy.sum(numpoly.monomial(3, dimensions=("q0", "q1")))
    with numpoly.global_options(**display_config):
        assert str(polynomial) == expected_output


def test_illegal_option():
    """Ensure that illegal arguments raises error."""
    with pytest.raises(KeyError):
        numpoly.set_options(not_an_argument=4)
    with pytest.raises(KeyError):
        with numpoly.global_options(not_an_argument=4):
            pass



"""Stack arrays in sequence depth wise (along third axis)."""
from __future__ import annotations
from typing import Sequence
import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.dstack)
def dstack(tup: Sequence[PolyLike]) -> ndpoly:
    """
    Stack arrays in sequence depth wise (along third axis).

    This is equivalent to concatenation along the third axis after 2-D arrays
    of shape `(M,N)` have been reshaped to `(M,N,1)` and 1-D arrays of shape
    `(N,)` have been reshaped to `(1,N,1)`. Rebuilds arrays divided by
    `dsplit`.

    This function makes most sense for arrays with up to 3 dimensions. For
    instance, for pixel-data with a height (first axis), width (second axis),
    and r/g/b channels (third axis). The functions `concatenate`, `stack` and
    `block` provide more general stacking and concatenation operations.

    Args:
        tup:
            The arrays must have the same shape along all but the third axis.
            1-D or 2-D arrays must have the same shape.

    Return:
        The array formed by stacking the given arrays, will be at least 3-D.

    Example:
        >>> poly1 = numpoly.variable(3)
        >>> const1 = numpoly.polynomial([1, 2, 3])
        >>> numpoly.dstack([poly1, const1])
        polynomial([[[q0, 1],
                     [q1, 2],
                     [q2, 3]]])
        >>> const2 = numpoly.polynomial([[1], [2], [3]])
        >>> poly2 = poly1.reshape(3, 1)
        >>> numpoly.dstack([const2, poly2])
        polynomial([[[1, q0]],
        <BLANKLINE>
                    [[2, q1]],
        <BLANKLINE>
                    [[3, q2]]])

    """
    arrays = numpoly.align_exponents(*tup)
    coefficients = [
        numpy.dstack([array.values[key] for array in arrays]) for key in arrays[0].keys
    ]
    return numpoly.polynomial_from_attributes(
        exponents=arrays[0].exponents,
        coefficients=coefficients,
        names=arrays[0].names,
        dtype=coefficients[0].dtype,
    )



"""View inputs as arrays with at least three dimensions."""
from __future__ import annotations
from typing import List, Union

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.atleast_3d)
def atleast_3d(*arys: PolyLike) -> Union[ndpoly, List[ndpoly]]:
    """
    View inputs as arrays with at least three dimensions.

    Args:
        arys:
            One or more array-like sequences. Non-array inputs are converted
            to arrays. Arrays that already have three or more dimensions are
            preserved.

    Return:
        An array, or list of arrays, each with ``a.ndim >= 3``.  Copies are
        avoided where possible, and views with three or more dimensions are
        returned.  For example, a 1-D array of shape ``(N,)`` becomes a view of
        shape ``(1, N, 1)``, and a 2-D array of shape ``(M, N)`` becomes a view
        of shape ``(M, N, 1)``.

    Example:
        >>> numpoly.atleast_3d(numpoly.variable())
        polynomial([[[q0]]])
        >>> a, b = numpoly.atleast_3d(1, [2, 3])
        >>> a
        polynomial([[[1]]])
        >>> b
        polynomial([[[2],
                     [3]]])

    """
    if len(arys) == 1:
        poly = numpoly.aspolynomial(arys[0])
        array = numpy.atleast_3d(poly.values)
        return numpoly.aspolynomial(array, names=poly.indeterminants)
    return [atleast_3d(ary) for ary in arys]



"""Calculate the absolute value element-wise."""
from __future__ import annotations
from typing import Any, Optional

import numpy
import numpy.typing

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements, simple_dispatch


@implements(numpy.absolute)
def absolute(
    x: PolyLike,
    out: Optional[ndpoly] = None,
    where: numpy.typing.ArrayLike = True,
    **kwargs: Any,
) -> ndpoly:
    r"""
    Calculate the absolute value element-wise.

    Args:
        x:
            Input array.
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Return:
        An ndarray containing the absolute value of. each element in `x`.
        For complex input, ``a + ib``, the absolute value is
        :math:`\sqrt{a^2+b^2}`. This is a scalar if `x` is a scalar.

    Example:
        >>> q0 = numpoly.variable()
        >>> poly = numpoly.polynomial([-1, 1, -2.3*q0, 2.3*q0])
        >>> poly
        polynomial([-1.0, 1.0, -2.3*q0, 2.3*q0])
        >>> numpoly.absolute(poly)
        polynomial([1.0, 1.0, 2.3*q0, 2.3*q0])
        >>> poly = numpoly.polynomial([q0, 1j*q0, (3+4j)*q0])
        >>> poly
        polynomial([q0, 1j*q0, (3+4j)*q0])
        >>> numpoly.absolute(poly)
        polynomial([q0, q0, 5.0*q0])

    """
    return simple_dispatch(
        numpy_func=numpy.absolute,
        inputs=(x,),
        out=None if out is None else (out,),
        where=where,
        **kwargs,
    )



"""Permute the dimensions of an array."""
from __future__ import annotations
from typing import Sequence, Union

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.transpose)
def transpose(
    a: PolyLike,
    axes: Union[None, Sequence[int], numpy.ndarray] = None,
) -> ndpoly:
    """
    Permute the dimensions of an array.

    Args:
        a:
            Input array.
        axes:
            By default, reverse the dimensions, otherwise permute the axes
            according to the values given.

    Return:
        `a` with its axes permuted. A view is returned whenever possible.

    Example:
        >>> poly = numpoly.monomial([3, 3]).reshape(2, 3)
        >>> poly
        polynomial([[1, q0, q0**2],
                    [q1, q0*q1, q1**2]])
        >>> numpoly.transpose(poly)
        polynomial([[1, q1],
                    [q0, q0*q1],
                    [q0**2, q1**2]])

    """
    a = numpoly.aspolynomial(a)
    out = numpy.transpose(a.values, axes=axes)
    out = numpoly.polynomial(out, names=a.indeterminants)
    return out



# cfrom_attributes.pyx

# Import Cython and types from standard Python
import numpy as np
cimport numpy as np
import numpoly
from typing import List


def cfrom_attributes(
    list coeffs,
    np.ndarray poly
):
    cdef Py_ssize_t i, nfields

    # Declare variables for string representation of polynomials
    names = poly.dtype.names
    nfields = len(names)

    for i in range(nfields):
        name = str(names[i])
        numpoly.cset_values(coeffs[i].ravel(), name, poly)




"""Testing saving to and loading from disk."""
from tempfile import TemporaryFile
import pickle
import pytest

import numpy
import numpoly


X, Y, Z = numpoly.variable(3)
ARRAY = numpy.array([1, 2, 3])
POLY = numpoly.polynomial([1, X, Z**2-1])


def test_savetxt(func_interface):
    outfile = TemporaryFile()
    func_interface.savetxt(outfile, POLY)
    outfile.seek(0)
    assert numpy.all(numpoly.loadtxt(outfile) == POLY)

    with open("/tmp/numpoly_save.npy", "wb") as dst:
        func_interface.savetxt(dst, POLY)
    with open("/tmp/numpoly_save.npy", "rb") as src:
        assert numpy.all(numpoly.loadtxt(src) == POLY)


def test_pickle():
    with open("/tmp/numpoly_pickle.pkl", "wb") as dst:
        pickle.dump(POLY, dst)
    with open("/tmp/numpoly_pickle.pkl", "rb") as src:
        assert numpy.all(pickle.load(src) == POLY)



"""Testing functions used for numpoly only functionality."""
from pytest import raises

import sympy
import numpy
import numpoly

X, Y = numpoly.symbols("q0"), numpoly.symbols("q1")
POLY1 = numpoly.polynomial([[1, X, X-1, X**2],
                            [Y, Y-1, Y**2, 1],
                            [X-1, X**2, 1, X],
                            [Y**2, 1, Y, Y-1]])


def test_call():
    poly = X+Y
    with raises(TypeError):
        poly(1, q0=2)
    with raises(TypeError):
        poly(1, 2, q1=3)
    with raises(TypeError):
        poly(not_an_arg=45)
    poly = numpoly.polynomial([2, X-Y+1])
    assert numpy.all(poly(q0=Y) == [2, 1])


def test_ndpoly():
    poly = numpoly.ndpoly(exponents=[(1,)], shape=(), names="q0")
    poly["<"] = 1
    assert poly == X
    poly = numpoly.ndpoly(exponents=[(1,)], shape=(), names=X)
    poly["<"] = 1
    assert poly == X
    poly = numpoly.ndpoly(
        exponents=[(1, 0), (0, 1)], shape=(), names=("q0" ,"q1"))
    poly["<;"] = 2
    poly[";<"] = 3
    assert poly == 2*X+3*Y
    with numpoly.global_options(varname_filter=r"Q\d+"):
        poly = numpoly.ndpoly(
            exponents=[(1, 0), (0, 1)], shape=(2,), names="Q:2")
        poly["<;"] = [1, 0]
        poly[";<"] = [0, 1]
        assert numpy.all(poly == numpoly.symbols("Q0 Q1"))


def test_poly_divide():
    const = numpoly.polynomial([1, 2])
    poly1 = numpoly.polynomial(X)
    poly2 = numpoly.polynomial([1, X])
    poly3 = numpoly.polynomial([[0, Y], [X, 1]])
    assert numpy.all(numpoly.poly_divide(poly1, const) == [X, 0.5*X])
    assert numpy.all(numpoly.poly_divide(const, poly1) == [0, 0])
    assert numpy.all(poly1.__truediv__(poly2) == [X, 1])
    assert numpy.all(poly2.__rtruediv__(poly3) == [[0, 0], [X, 0]])
    assert numpy.all(poly2.__div__(poly1) == [0, 1])
    assert numpy.all(poly3.__rdiv__(poly2) == [[0, 0], [0, X]])


def test_poly_divmod():
    assert numpoly.poly_divmod(X+3, X+2) == (1, 1)
    assert numpoly.poly_divmod(Y+3, X+2) == (0, Y+3)
    assert numpoly.poly_divmod(3, X+2) == (0, 3)
    assert numpoly.poly_divmod(X, X+2) == (1, -2)
    assert numpoly.poly_divmod(Y, X+2) == (0, Y)
    assert numpoly.poly_divmod(X*Y, X+2) == (Y, -2*Y)

    x1 = numpoly.polynomial(-1.715408531156317e+18)
    x2 = numpoly.polynomial(3.097893826691672e+18)
    divided, remainder = numpoly.poly_divmod(x1, x2)
    assert numpoly.allclose(x1, x2*divided)
    assert remainder == 0

    # No candidate and no infinity loops when cutoff is reached:
    assert numpoly.poly_function.get_division_candidate(x1, x2, cutoff=1) is None

    x1 = (24*3600)**2*398600.4415
    x2 = 2.689498059130061e-63*X**2+4.351432444850692e-27*X+1760083471.506941
    divided, remainder = numpoly.poly_divmod(x1, x2)
    assert numpoly.allclose(x1, x2*divided+remainder)
    assert numpy.allclose(remainder, x1)

    assert divmod(X+3, 2) == (0.5*X+1.5, 0)
    assert divmod(Y+3, 2) == (0.5*Y+1.5, 0)
    assert divmod(3, numpoly.polynomial(2)) == (1.5, 0)
    assert divmod(X, 2) == (0.5*X, 0)
    assert divmod(Y, 2) == (0.5*Y, 0)
    assert divmod(X*Y, 2) == (0.5*X*Y, 0)

    assert (X+3).__divmod__(X) == (1, 3)
    assert (Y+3).__divmod__(X) == (0, Y+3)
    with raises(numpoly.FeatureNotSupported):
        assert numpy.array(3).__divmod__(X) == (0, 3)
    assert X.__divmod__(X) == (1, 0)
    assert Y.__divmod__(X) == (0, Y)
    assert (X*Y).__divmod__(X) == (Y, 0)


def test_isconstant():
    assert not numpoly.polynomial(X).isconstant()
    assert numpoly.polynomial(1).isconstant()
    assert not numpoly.polynomial([1, X]).isconstant()
    assert numpoly.polynomial([1, 2]).isconstant()


def test_tonumpy():
    assert isinstance(numpoly.tonumpy(numpoly.polynomial([1, 2, 3])), numpy.ndarray)
    with raises(numpoly.FeatureNotSupported):
        numpoly.tonumpy(X)


def test_poly_remainder():
    poly = numpoly.polynomial([[1, 2*X], [3*Y+X, 4]])
    assert numpy.all(numpoly.poly_remainder(poly, 2) == [[0, 0], [0, 0]])
    assert numpy.all(numpy.array([1, 2]) % poly == [[0, 2], [1, 0]])
    assert numpy.all(poly % numpy.array([1, 2]) == 0)
    assert numpy.all(poly.__mod__(numpy.array([1, 2])) == 0)
    assert numpy.all(poly.__rmod__(numpy.array([1, 2])) == [[0, 2], [1, 0]])


def test_sortable_proxy():
    assert numpy.all(numpoly.sortable_proxy(numpy.arange(4)) == [0, 1, 2, 3])
    assert numpy.all(numpoly.sortable_proxy(-numpy.arange(4)) == [3, 2, 1, 0])

    assert numpy.all(numpoly.sortable_proxy(numpy.arange(4)*X) == [0, 1, 2, 3])
    assert numpy.all(numpoly.sortable_proxy(-numpy.arange(4)*X) == [0, 3, 2, 1])
    assert numpy.all(numpoly.sortable_proxy(X**numpy.arange(4)) == [0, 1, 2, 3])
    assert numpy.all(numpoly.sortable_proxy(-X**numpy.arange(4)) == [0, 1, 2, 3])
    assert numpy.all(numpoly.sortable_proxy(X**4+X**numpy.arange(4)) == [0, 1, 2, 3])
    assert numpy.all(numpoly.sortable_proxy(X**4-X**numpy.arange(4)) == [0, 1, 2, 3])

    poly = numpoly.polynomial([1, X, Y, X**2, X*Y, Y**2])

    proxy = numpoly.sortable_proxy(poly, graded=False, reverse=False)
    assert numpy.all(poly[numpy.argsort(proxy)] == [1, X, X**2, Y, X*Y, Y**2])
    proxy = numpoly.sortable_proxy(poly, graded=True, reverse=False)
    assert numpy.all(poly[numpy.argsort(proxy)] == [1, X, Y, X**2, X*Y, Y**2])
    proxy = numpoly.sortable_proxy(poly, graded=True, reverse=True)
    assert numpy.all(poly[numpy.argsort(proxy)] == [1, Y, X, Y**2, X*Y, X**2])
    proxy = numpoly.sortable_proxy(poly, graded=False, reverse=True)
    assert numpy.all(poly[numpy.argsort(proxy)] == [1, Y, Y**2, X, X*Y, X**2])

    assert numpy.all(numpoly.sortable_proxy(POLY1) ==
                     [[ 0,  4,  5,  8],
                      [10, 11, 14,  1],
                      [ 6,  9,  2,  7],
                      [15,  3, 12, 13]])
    assert numpy.all(numpoly.sortable_proxy(POLY1, graded=True) ==
                     [[ 0,  4,  5, 12],
                      [ 8,  9, 14,  1],
                      [ 6, 13,  2,  7],
                      [15,  3, 10, 11]])
    assert numpy.all(numpoly.sortable_proxy(POLY1, reverse=True) ==
                     [[ 0, 10, 11, 14],
                      [ 4,  5,  8,  1],
                      [12, 15,  2, 13],
                      [ 9,  3,  6,  7]])


def test_lead_exponent():
    empty = numpoly.lead_exponent([])
    assert empty.size == 0
    assert empty.dtype == int
    assert empty.shape == (0, 1)
    exponents = numpoly.lead_exponent(POLY1)
    xs, ys = exponents[:, :, 0], exponents[:, :, 1]
    assert numpy.all(xs == [[0, 1, 1, 2], [0, 0, 0, 0],
                            [1, 2, 0, 1], [0, 0, 0, 0]])
    assert numpy.all(ys == [[0, 0, 0, 0], [1, 1, 2, 0],
                            [0, 0, 0, 0], [2, 0, 1, 1]])


def test_lead_coefficient():
   empty = numpoly.lead_coefficient([])
   assert empty.size == 0
   assert empty.dtype == int
   assert empty.shape == (0,)


def test_set_dimensions():
    """Tests for numpoly.set_dimensions."""
    assert numpoly.set_dimensions(POLY1).names == ("q0", "q1", "q2")
    assert numpoly.set_dimensions(POLY1, 1).names == ("q0",)
    assert numpoly.set_dimensions(POLY1, 2).names == ("q0", "q1")
    assert numpoly.set_dimensions(POLY1, 3).names == ("q0", "q1", "q2")
    assert numpoly.set_dimensions(POLY1, 4).names == ("q0", "q1", "q2", "q3")



"""Test ndpoly baseclass functionality."""
import numpy
from pytest import raises
import numpoly

XY = numpoly.variable(2)
X, Y = numpoly.symbols("q0"), numpoly.symbols("q1")
EMPTY = numpoly.polynomial([])


def test_scalars():
    """Test scalar objects to catch edgecases."""
    assert XY.shape == (2,)
    assert XY.size == 2
    assert X.shape == ()
    assert X.size == 1
    assert EMPTY.shape in [(), (0,)]  # different behavior in py2/3
    assert EMPTY.size == 0

    assert numpy.all(numpy.array(XY.coefficients) == [[1, 0], [0, 1]])
    assert X.coefficients == [1]
    assert EMPTY.coefficients == []

    assert numpy.all(XY.exponents == [[1, 0], [0, 1]])
    assert XY.exponents.shape == (2, 2)
    assert X.exponents == 1
    assert X.exponents.shape == (1, 1)
    assert numpy.all(EMPTY.exponents == 0)
    assert EMPTY.exponents.shape == (1, 1)

    assert numpy.all(XY.indeterminants == XY)
    assert X.indeterminants == X

    assert numpy.all(XY.values == numpy.array(
        [(1, 0), (0, 1)], dtype=[("<;", int), (";<", int)]))
    assert X.values == numpy.array((1,), dtype=[("<", int)])
    assert EMPTY.values.dtype == numpy.dtype([(";", int)])
    assert not EMPTY.values.size

    assert not X.isconstant()
    assert EMPTY.isconstant()

    assert X.todict() == {(1,): 1}
    assert EMPTY.todict() == {}

    assert isinstance(EMPTY.tonumpy(), numpy.ndarray)

    assert X.dtype == int
    assert X.astype(float).dtype == float
    assert EMPTY.dtype == int
    assert EMPTY.astype(float).dtype == float


def test_dispatch_array_ufunc():
    """Test dispatch for ufuncs."""
    assert numpoly.sum(XY) == XY.__array_ufunc__(numpy.sum, "__call__", XY)
    with raises(numpoly.FeatureNotSupported):
        XY.__array_ufunc__(numpy.sum, "not_a_method", XY)
    with raises(numpoly.FeatureNotSupported):
        XY.__array_ufunc__(object, "__call__", XY)


def test_dispatch_array_function():
    """Test dispatch for functions."""
    assert numpoly.sum(XY) == XY.__array_function__(numpy.sum, (int,), (XY,), {})
    with raises(numpoly.FeatureNotSupported):
        XY.__array_function__(object, (int,), (XY,), {})



"""Cast polynomial to numpy.ndarray, if possible."""
from __future__ import annotations

import numpy
import numpoly

from ..baseclass import PolyLike


def tonumpy(poly: PolyLike) -> numpy.ndarray:
    """
    Cast polynomial to numpy.ndarray, if possible.

    Args:
        poly:
            polynomial to cast.

    Return:
        Numpy array.

    Raise:
        numpoly.baseclass.FeatureNotSupported:
            Only constant polynomials can be cast to numpy.ndarray.

    Example:
        >>> numpoly.tonumpy(numpoly.polynomial([1, 2]))
        array([1, 2])

    """
    poly = numpoly.aspolynomial(poly)
    if not poly.isconstant():
        raise numpoly.FeatureNotSupported(
            "only constant polynomials can be converted to array."
        )
    idx = numpy.argwhere(numpy.all(poly.exponents == 0, -1)).item()
    if poly.size:
        return numpy.array(poly.coefficients[idx])
    return numpy.array([])



"""Polynomial functionality."""
from .call import call
from .decompose import decompose
from .derivative import derivative, gradient, hessian
from .isconstant import isconstant
from .lead_coefficient import lead_coefficient
from .lead_exponent import lead_exponent
from .divide import get_division_candidate, poly_divide, poly_divmod, poly_remainder
from .set_dimensions import set_dimensions
from .sortable_proxy import sortable_proxy
from .tonumpy import tonumpy

__all__ = (
    "call",
    "decompose",
    "derivative",
    "gradient",
    "hessian",
    "isconstant",
    "lead_coefficient",
    "lead_exponent",
    "poly_divide",
    "poly_divmod",
    "poly_remainder",
    "set_dimensions",
    "sortable_proxy",
    "tonumpy",
)



"""Construct symbol variables."""
from typing import Optional, Sequence
import re

import numpy
import numpy.typing

import numpoly
from ..baseclass import ndpoly


def symbols(
    names: Optional[Sequence[str]] = None,
    asarray: bool = False,
    dtype: numpy.typing.DTypeLike = "i8",
    allocation: Optional[int] = None,
) -> ndpoly:
    """
    Construct symbol variables.

    Most directly be providing a list of string names. But a set of shorthands
    also exists:

    * ``,`` and `` `` (space) can be used as a variable delimiter.
    * ``{number}:{number}`` can be used to define a numerical range.
    * ``{letter}:{letter}`` can be used to define a alphabet range.

    Args:
        names:
            Indeterminants are determined by splitting the string on space. If
            iterable of strings, indeterminants defined directly.
        asarray:
            Enforce output as array even in the case where there is only one
            variable.
        dtype:
            The data type of the polynomial coefficients.
        allocation:
            The maximum number of polynomial exponents. If omitted, use
            length of exponents for allocation.

    Return:
        Polynomial array with unit components in each dimension.

    Example:
        >>> numpoly.symbols()
        polynomial(q0)
        >>> numpoly.symbols("q4")
        polynomial(q4)
        >>> numpoly.symbols("q:7")
        polynomial([q0, q1, q2, q3, q4, q5, q6])
        >>> numpoly.symbols("q3:6")
        polynomial([q3, q4, q5])
        >>> numpoly.symbols(["q0", "q3", "q99"])
        polynomial([q0, q3, q99])

    """
    if names is None:
        coefficients = [numpy.ones(1, dtype=dtype)]
        out = numpoly.ndpoly.from_attributes(
            exponents=[(1,)],
            coefficients=coefficients,
            dtype=dtype,
            allocation=allocation,
        )
        if not asarray:
            out = numpoly.aspolynomial(out[0])
        return out

    if not isinstance(names, str):
        names = tuple(names)

    else:
        names = re.sub(" ", ",", names)
        if "," in names:
            asarray = True
            names = tuple(name for name in names.split(",") if name)

        else:
            match = re.search(r"(\d*):(\d+)", names)
            if match:
                start = int(match.group(1) or 0)
                end = int(match.group(2))
                names = tuple(
                    names.replace(match.group(0), str(idx)) for idx in range(start, end)
                )

            else:
                names = (names,)

    assert isinstance(names, tuple)
    exponents = numpy.eye(len(names), dtype=int)
    coefficients = numpy.eye(len(names), dtype=dtype)
    out = numpoly.ndpoly.from_attributes(
        exponents=exponents,
        coefficients=coefficients,
        names=names,
        allocation=allocation,
    )
    if out.size == 1 and not asarray:
        return numpoly.aspolynomial(out[0])
    return out



"""Global numpoly options."""
from typing import Any, Dict, Iterator
from contextlib import contextmanager

GLOBAL_OPTIONS_DEFAULTS = {
    "default_varname": "q",
    "display_graded": True,
    "display_reverse": False,
    "display_inverse": True,
    "display_exponent": "**",
    "display_multiply": "*",
    "force_number_suffix": True,
    "retain_names": True,
    "retain_coefficients": False,
    "sort_graded": True,
    "sort_reverse": False,
    "varname_filter": r"q\d+",
}
_NUMPOLY_OPTIONS = GLOBAL_OPTIONS_DEFAULTS.copy()


def get_options(defaults: bool = False) -> Dict[str, Any]:
    """
    Get global numpoly options.

    Args:
        defaults:
            Return the options for the global option defaults, instead of the
            options currently in use.

    Return:
        Collection of values that affect the numpoly machinery.
        See `numpoly.set_options` for details.

    Example:
        >>> options = get_options()
        >>> options["default_varname"]
        'q'

    Note:
        `set_options`, `global_options`

    """
    if defaults:
        return GLOBAL_OPTIONS_DEFAULTS.copy()
    return _NUMPOLY_OPTIONS.copy()


def set_options(**kwargs: Any) -> None:
    """
    Set global numpoly options.

    Args:
        default_varname:
            Polynomial indeterminant defaults, if not defined explicitly.
        display_graded:
            When displaying polynomials as strings, sort polynomial sums in
            graded order.
        display_reverse:
            When displaying polynomials as strings, sort polynomial sums in
            reversed lexicographical order.
        display_inverse:
            If true, display polynomials from smallest to largest exponent.
        display_exponent:
            Exponent sign; Separate indeterminants and its power.
        display_multiply:
            Multiplication sign; Separates coefficients and
            indeterminants, and indeterminants from each other.
        force_number_suffix:
            Add a postfix index to single indeterminant name. If single
            indeterminant name, e.g. 'q' is provided, but the polynomial is
            multivariate, an extra postfix index is added to differentiate the
            names: 'q0, q1, q2, ...'. If true, enforce this behavior for single
            variables as well such that 'q' always get converted to 'q0'.
        retain_coefficients:
            After each operation a cleanup is done to reduce the polynomial to
            its smallest memory imprint. If true, do not remove redundant
            coefficients (consisting of only zeros).
        retain_names:
            After each operation a cleanup is done to reduce the polynomial to
            its smallest memory imprint. If true, do not remove redundant
            names (not represented in polynomial anymore).
        sort_graded:
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``x**2*y**2*z**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``x**3*y*z``, ``x*y**2*z and
            ``x*y*z**2``, which all have exponent sum of 5.
        sort_reverse:
            Reverse lexicographical sorting meaning that ``x*y**3`` is
            considered bigger than ``x**3*y``, instead of the opposite.
        varname_filter:
            Regular expression defining valid indeterminant names.

    Example:
        >>> numpoly.monomial([3, 3])
        polynomial([1, q0, q0**2, q1, q0*q1, q1**2])
        >>> numpoly.set_options(default_varname="z", varname_filter=".+")
        >>> numpoly.monomial([3, 3])
        polynomial([1, z0, z0**2, z1, z0*z1, z1**2])

    Note:
        `get_options`, `global_options`

    """
    for key in kwargs:
        if key not in _NUMPOLY_OPTIONS:
            raise KeyError(f"option '{key}' not recognized.")
    _NUMPOLY_OPTIONS.update(**kwargs)


@contextmanager
def global_options(**kwargs: Any) -> Iterator[Dict[str, Any]]:
    """
    Temporarily set global numpoly options.

    Args:
        kwargs:
            Collection of values that deviate from the defaults.
            See `numpoly.set_options` for details.

    Yield:
        Collection of values that affect the numpoly machinery.
        See `numpoly.set_options` for details.

    Example:
        >>> numpoly.get_options()["default_varname"]
        'q'
        >>> with numpoly.global_options(default_varname="X"):
        ...     print(numpoly.get_options()["default_varname"])
        X
        >>> numpoly.get_options()["default_varname"]
        'q'

    Note:
        `get_options`, `set_options`

    """
    options = get_options()
    set_options(**kwargs)
    try:
        yield get_options()
    finally:
        set_options(**options)



"""Construct polynomial from polynomial attributes."""
from __future__ import annotations
from typing import Optional, Sequence, Tuple, Union

import numpy.typing

import numpoly

from . import clean
from ..baseclass import ndpoly


def polynomial_from_attributes(
    exponents: numpy.typing.ArrayLike,
    coefficients: Sequence[numpy.typing.ArrayLike],
    names: Union[None, str, Tuple[str, ...], ndpoly] = None,
    dtype: Optional[numpy.typing.DTypeLike] = None,
    allocation: Optional[int] = None,
    retain_coefficients: Optional[bool] = None,
    retain_names: Optional[bool] = None,
) -> ndpoly:
    """
    Construct polynomial from polynomial attributes.

    Args:
        exponents:
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        coefficients:
            The polynomial coefficients. Must correspond to `exponents` by
            having the same length ``N``.
        names:
            The indeterminant names, either as string names or as
            simple polynomials. Must correspond to the exponents by having
            length ``D``.
        dtype:
            The data type of the polynomial. If omitted, extract from
            `coefficients`.
        allocation:
            The maximum number of polynomial exponents. If omitted, use
            length of exponents for allocation.
        retain_coefficients:
            Do not remove redundant coefficients. If omitted use global
            defaults.
        retain_names:
            Do not remove redundant names. If omitted use global defaults.

    Return:
        Polynomial array with attributes determined by the input.

    Example:
        >>> numpoly.ndpoly.from_attributes(
        ...     exponents=[(0,), (1,)],
        ...     coefficients=[[1, 0], [0, 1]],
        ...     names="q4",
        ... )
        polynomial([1, q4])
        >>> numpoly.ndpoly.from_attributes(
        ...     exponents=[(0, 0, 0), (1, 1, 2)],
        ...     coefficients=[4, -1],
        ...     names=("q2", "q4", "q10"),
        ... )
        polynomial(-q2*q4*q10**2+4)
        >>> numpoly.ndpoly.from_attributes(
        ...     exponents=[(0,)],
        ...     coefficients=[0],
        ... )
        polynomial(0)

    """
    exponents, coefficients, names = clean.postprocess_attributes(
        exponents=exponents,
        coefficients=coefficients,
        names=names,
        retain_coefficients=retain_coefficients,
        retain_names=retain_names,
    )
    if coefficients:
        dtype = coefficients[0].dtype if dtype is None else dtype
        shape = coefficients[0].shape
    else:
        dtype = dtype if dtype else int
        shape = ()

    poly = numpoly.ndpoly(
        exponents=exponents,
        shape=shape,
        names=names,
        dtype=dtype,
        allocation=allocation,
    )

    if coefficients:
        numpoly.cfrom_attributes(coefficients, poly.values.ravel())

    # for key, values in zip(poly.keys, coefficients):
    #    poly.values[key] = values

    return poly



"""Stack arrays in sequence vertically (row wise)."""
from __future__ import annotations
from typing import Sequence

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.vstack)
def vstack(tup: Sequence[PolyLike]) -> ndpoly:
    """
    Stack arrays in sequence vertically (row wise).

    This is equivalent to concatenation along the first axis after 1-D arrays
    of shape `(N,)` have been reshaped to `(1,N)`. Rebuilds arrays divided by
    `vsplit`.

    This function makes most sense for arrays with up to 3 dimensions. For
    instance, for pixel-data with a height (first axis), width (second axis),
    and r/g/b channels (third axis). The functions `concatenate`, `stack` and
    `block` provide more general stacking and concatenation operations.

    Args:
        tup:
            The arrays must have the same shape along all but the first axis.
            1-D arrays must have the same length.

    Return:
        The array formed by stacking the given arrays, will be at least 2-D.

    Example:
        >>> poly1 = numpoly.variable(3)
        >>> const1 = numpoly.polynomial([1, 2, 3])
        >>> numpoly.vstack([poly1, const1])
        polynomial([[q0, q1, q2],
                    [1, 2, 3]])
        >>> const2 = numpoly.polynomial([[1], [2], [3]])
        >>> poly2 = poly1.reshape(3, 1)
        >>> numpoly.vstack([const2, poly2])
        polynomial([[1],
                    [2],
                    [3],
                    [q0],
                    [q1],
                    [q2]])

    """
    arrays = numpoly.align_exponents(*tup)
    coefficients = [
        numpy.vstack([array.values[key] for array in arrays]) for key in arrays[0].keys
    ]
    return numpoly.polynomial_from_attributes(
        exponents=arrays[0].exponents,
        coefficients=coefficients,
        names=arrays[0].names,
        dtype=coefficients[0].dtype,
    )



"""Counts the number of non-zero values in the array a."""
from __future__ import annotations
from typing import Any, Sequence, Union

import numpy
import numpy.typing
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


@implements(numpy.count_nonzero)
def count_nonzero(
    x: PolyLike,
    axis: Union[None, int, Sequence[int]] = None,
    **kwargs: Any,
) -> Union[int, numpy.ndarray]:
    """
    Count the number of non-zero values in the array a.

    Args:
        x:
            The array for which to count non-zeros.
        axis: (Union[int, Tuple[int], None]):
            Axis or tuple of axes along which to count non-zeros. Default is
            None, meaning that non-zeros will be counted along a flattened
            version of a.

    Return:
        Number of non-zero values in the array along a given axis. Otherwise,
        the total number of non-zero values in the array is returned.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.count_nonzero([q0])
        1
        >>> numpoly.count_nonzero([[0, q0*q0, 0, 0],
        ...                        [q0+1, 0, 2*q0, 7*q0]])
        4
        >>> numpoly.count_nonzero([[0, q0+q1, 0, 0],
        ...                        [3*q1, 0, 2, 7*q0]], axis=0)
        array([1, 1, 1, 1])
        >>> numpoly.count_nonzero([[0, q1, 0, 0],
        ...                        [q0, 0, 2*q0, 6*q1]], axis=1)
        array([1, 3])

    """
    a = numpoly.aspolynomial(x)
    index = numpy.any(numpy.asarray(a.coefficients), axis=0)
    return numpy.count_nonzero(index, axis=axis)



"""Align polynomials."""
from __future__ import annotations
from typing import Tuple

import numpy
import numpoly
from .baseclass import ndpoly, PolyLike
from .option import get_options


def align_polynomials(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomial such that dimensionality, shape, etc. are compatible.

    Alignment includes shape (for broadcasting), indeterminants, exponents and
    dtype.

    Args:
        polys:
            Polynomial to make adjustment to.

    Return:
        Same as ``polys``, but internal adjustments made to make them
        compatible for further operations.

    Example:
        >>> q0 = numpoly.variable()
        >>> q0q1 = numpoly.variable(2)
        >>> q0
        polynomial(q0)
        >>> q0.coefficients
        [np.int64(1)]
        >>> q0.indeterminants
        polynomial([q0])
        >>> q0, _ = numpoly.align_polynomials(q0, q0q1)
        >>> q0
        polynomial([q0, q0])
        >>> q0.coefficients
        [array([0, 0]), array([1, 1])]
        >>> q0.indeterminants
        polynomial([q0, q1])

    """
    # polys = numpoly.broadcast_arrays(*polys)
    polys = align_shape(*polys)
    # polys = align_indeterminants(*polys)
    polys = align_exponents(*polys)
    return polys


def align_shape(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomial by shape.

    Args:
        polys:
            Polynomial to make adjustment to.

    Return:
        Same as ``polys``, but internal adjustments made to make them
        compatible for further operations.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly1 = 4*q0
        >>> poly2 = numpoly.polynomial([[2*q0+1, 3*q0-q1]])
        >>> poly1.shape
        ()
        >>> poly2.shape
        (1, 2)
        >>> poly1, poly2 = numpoly.align_shape(poly1, poly2)
        >>> poly1
        polynomial([[4*q0, 4*q0]])
        >>> poly2
        polynomial([[2*q0+1, -q1+3*q0]])
        >>> poly1.shape
        (1, 2)
        >>> poly2.shape
        (1, 2)

    """
    # return tuple(numpoly.broadcast_arrays(*polys))
    polys_ = [numpoly.aspolynomial(poly) for poly in polys]
    common = numpy.ones(
        numpy.broadcast_shapes(*[poly.shape for poly in polys_]), dtype=int
    )

    for idx, poly in enumerate(polys_):
        if poly.shape != common.shape:
            polys_[idx] = poly.from_attributes(
                exponents=poly.exponents,
                coefficients=tuple(coeff * common for coeff in poly.coefficients),
                names=poly.indeterminants,
            )
    return tuple(polys_)


def align_indeterminants(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomial by indeterminants.

    Args:
        polys:
            Polynomial to make adjustment to.

    Return:
        Same as ``polys``, but internal adjustments made to make them
        compatible for further operations.

    Example:
        >>> q0 = numpoly.variable()
        >>> poly1 = numpoly.polynomial(2*q0+1)
        >>> q0, q1 = numpoly.variable(2)
        >>> poly2 = numpoly.polynomial(3*q0-q1)
        >>> poly1.indeterminants
        polynomial([q0])
        >>> poly2.indeterminants
        polynomial([q0, q1])
        >>> poly1, poly2 = numpoly.align_indeterminants(poly1, poly2)
        >>> poly1
        polynomial(2*q0+1)
        >>> poly2
        polynomial(-q1+3*q0)
        >>> poly1.indeterminants
        polynomial([q0, q1])
        >>> poly2.indeterminants
        polynomial([q0, q1])

    """
    polys_ = [numpoly.aspolynomial(poly) for poly in polys]

    options = get_options()
    length = len(options["default_varname"])
    common_names = tuple(
        sorted(
            {str(name) for poly in polys_ for name in poly.names},
            key=lambda x: int(x[length:] or "0"),
        )
    )
    if not common_names:
        return tuple(polys_)

    for idx, poly in enumerate(polys_):
        if poly.names == common_names:
            continue
        indices = numpy.array(
            [common_names.index(name) for name in poly.names if name in common_names]
        )
        exponents = numpy.zeros((len(poly.keys), len(common_names)), dtype=int)
        if indices.size:
            exponents[:, indices] = poly.exponents
        polys_[idx] = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=poly.coefficients,
            names=common_names,
            retain_coefficients=True,
            retain_names=True,
        )
    return tuple(polys_)


def align_exponents(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomials such that the exponents are the same.

    Aligning exponents assumes that the indeterminants is also aligned.

    Args:
        polys:
            Polynomial to make adjustment to.

    Return:
        Same as ``polys``, but internal adjustments made to make them
        compatible for further operations.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly1 = q0*q1
        >>> poly2 = numpoly.polynomial([q0**5, q1**3-1])
        >>> poly1.exponents
        array([[1, 1]], dtype=uint32)
        >>> poly2.exponents
        array([[0, 0],
               [0, 3],
               [5, 0]], dtype=uint32)
        >>> poly1, poly2 = numpoly.align_exponents(poly1, poly2)
        >>> poly1
        polynomial(q0*q1)
        >>> poly2
        polynomial([q0**5, q1**3-1])
        >>> poly1.exponents
        array([[0, 0],
               [0, 3],
               [1, 1],
               [5, 0]], dtype=uint32)
        >>> poly2.exponents
        array([[0, 0],
               [0, 3],
               [1, 1],
               [5, 0]], dtype=uint32)

    """
    polys_ = [numpoly.aspolynomial(poly) for poly in polys]
    if not all(polys_[0].names == poly.names for poly in polys_):
        polys_ = list(align_indeterminants(*polys_))

    global_exponents = numpy.vstack([poly.exponents for poly in polys_])
    global_exponents = numpy.unique(global_exponents, axis=0).tolist()

    for idx, poly in enumerate(polys_):
        lookup = {
            tuple(exponent): coefficient
            for exponent, coefficient in zip(poly.exponents, poly.coefficients)
        }

        zeros = numpy.zeros(poly.shape, dtype=poly.dtype)
        coefficients = [
            lookup.get(tuple(exponent), zeros) for exponent in global_exponents
        ]
        polys_[idx] = poly.from_attributes(
            exponents=global_exponents,
            coefficients=coefficients,
            names=poly.names,
            retain_coefficients=True,
            retain_names=True,
        )
    return tuple(polys_)
